@@ -1,6 +1,6 @@
 /*
     SDL - Simple DirectMedia Layer
-    Copyright (C) 1997, 1998, 1999, 2000, 2001, 2002  Sam Lantinga
+    Copyright (C) 1997-2004 Sam Lantinga
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Library General Public
@@ -22,7 +22,7 @@
 
 #ifdef SAVE_RCSID
 static char rcsid =
- "@(#) $Id: SDL_dx5video.c,v 1.10 2002/04/11 15:23:07 slouken Exp $";
+ "@(#) $Id: SDL_dx5video.c,v 1.26 2004/02/16 21:09:24 slouken Exp $";
 #endif
 
 #include <stdio.h>
@@ -41,7 +41,6 @@ static char rcsid =
 #include "SDL_events.h"
 #include "SDL_syswm.h"
 #include "SDL_sysvideo.h"
-#include "SDL_RLEaccel_c.h"
 #include "SDL_blit.h"
 #include "SDL_pixels_c.h"
 #include "SDL_dx5video.h"
@@ -51,14 +50,28 @@ static char rcsid =
 #include "SDL_dx5yuv_c.h"
 #include "SDL_wingl_c.h"
 
+#ifdef _WIN32_WCE
+#define NO_CHANGEDISPLAYSETTINGS
+#endif
+#ifndef WS_MAXIMIZE
+#define WS_MAXIMIZE		0
+#endif
+#ifndef SWP_NOCOPYBITS
+#define SWP_NOCOPYBITS	0
+#endif
+#ifndef PC_NOCOLLAPSE
+#define PC_NOCOLLAPSE	0
+#endif
+
 
 /* DirectX function pointers for video and events */
 HRESULT (WINAPI *DDrawCreate)( GUID FAR *lpGUID, LPDIRECTDRAW FAR *lplpDD, IUnknown FAR *pUnkOuter );
-HRESULT (WINAPI *DInputCreate)(HINSTANCE hinst, DWORD dwVersion, LPDIRECTINPUTA *ppDI, LPUNKNOWN punkOuter);
+HRESULT (WINAPI *DInputCreate)(HINSTANCE hinst, DWORD dwVersion, LPDIRECTINPUT *ppDI, LPUNKNOWN punkOuter);
 
 /* This is the rect EnumModes2 uses */
 struct DX5EnumRect {
 	SDL_Rect r;
+	int refreshRate;
 	struct DX5EnumRect* next;
 };
 static struct DX5EnumRect *enumlists[NUM_MODELISTS];
@@ -443,19 +456,19 @@ static int DX5_Available(void)
 
 	/* Version check DINPUT.DLL and DDRAW.DLL (Is DirectX okay?) */
 	dinput_ok = 0;
-	DInputDLL = LoadLibrary("DINPUT.DLL");
+	DInputDLL = LoadLibrary(TEXT("DINPUT.DLL"));
 	if ( DInputDLL != NULL ) {
 		dinput_ok = 1;
 	  	FreeLibrary(DInputDLL);
 	}
 	ddraw_ok = 0;
-	DDrawDLL = LoadLibrary("DDRAW.DLL");
+	DDrawDLL = LoadLibrary(TEXT("DDRAW.DLL"));
 	if ( DDrawDLL != NULL ) {
 	  HRESULT (WINAPI *DDrawCreate)(GUID *,LPDIRECTDRAW *,IUnknown *);
 	  LPDIRECTDRAW DDraw;
 
 	  /* Try to create a valid DirectDraw object */
-	  DDrawCreate = (void *)GetProcAddress(DDrawDLL, "DirectDrawCreate");
+	  DDrawCreate = (void *)GetProcAddress(DDrawDLL, TEXT("DirectDrawCreate"));
 	  if ( (DDrawCreate != NULL)
 			&& !FAILED(DDrawCreate(NULL, &DDraw, NULL)) ) {
 	    if ( !FAILED(IDirectDraw_SetCooperativeLevel(DDraw,
@@ -511,15 +524,15 @@ static int DX5_Load(void)
 	int status;
 
 	DX5_Unload();
-	DDrawDLL = LoadLibrary("DDRAW.DLL");
+	DDrawDLL = LoadLibrary(TEXT("DDRAW.DLL"));
 	if ( DDrawDLL != NULL ) {
 		DDrawCreate = (void *)GetProcAddress(DDrawDLL,
-					"DirectDrawCreate");
+					TEXT("DirectDrawCreate"));
 	}
-	DInputDLL = LoadLibrary("DINPUT.DLL");
+	DInputDLL = LoadLibrary(TEXT("DINPUT.DLL"));
 	if ( DInputDLL != NULL ) {
 		DInputCreate = (void *)GetProcAddress(DInputDLL,
-					"DirectInputCreateA");
+					TEXT("DirectInputCreateA"));
 	}
 	if ( DDrawDLL && DDrawCreate && DInputDLL && DInputCreate ) {
 		status = 0;
@@ -596,11 +609,11 @@ static SDL_VideoDevice *DX5_CreateDevice(int devindex)
 	device->SetGammaRamp = DX5_SetGammaRamp;
 	device->GetGammaRamp = DX5_GetGammaRamp;
 #ifdef HAVE_OPENGL
-        device->GL_LoadLibrary = WIN_GL_LoadLibrary;
-        device->GL_GetProcAddress = WIN_GL_GetProcAddress;
-        device->GL_GetAttribute = WIN_GL_GetAttribute;
-        device->GL_MakeCurrent = WIN_GL_MakeCurrent;
-        device->GL_SwapBuffers = WIN_GL_SwapBuffers;
+	device->GL_LoadLibrary = WIN_GL_LoadLibrary;
+	device->GL_GetProcAddress = WIN_GL_GetProcAddress;
+	device->GL_GetAttribute = WIN_GL_GetAttribute;
+	device->GL_MakeCurrent = WIN_GL_MakeCurrent;
+	device->GL_SwapBuffers = WIN_GL_SwapBuffers;
 #endif
 	device->SetCaption = WIN_SetWMCaption;
 	device->SetIcon = WIN_SetWMIcon;
@@ -638,8 +651,10 @@ static HRESULT WINAPI EnumModes2(DDSURFACEDESC *desc, VOID *udata)
 	struct DX5EnumRect *enumrect;
 #if defined(NONAMELESSUNION)
 	int bpp = desc->ddpfPixelFormat.u1.dwRGBBitCount;
+	int refreshRate = desc->u2.dwRefreshRate;
 #else
 	int bpp = desc->ddpfPixelFormat.dwRGBBitCount;
+	int refreshRate = desc->dwRefreshRate;
 #endif
 
 	switch (bpp)  {
@@ -648,29 +663,44 @@ static HRESULT WINAPI EnumModes2(DDSURFACEDESC *desc, VOID *udata)
 		case 24:
 		case 32:
 			bpp /= 8; --bpp;
+			if ( enumlists[bpp] &&
+			     enumlists[bpp]->r.w == (Uint16)desc->dwWidth &&
+			     enumlists[bpp]->r.h == (Uint16)desc->dwHeight ) {
+				if ( refreshRate > enumlists[bpp]->refreshRate &&
+				     refreshRate <= 85 /* safe value? */ ) {
+					enumlists[bpp]->refreshRate = refreshRate;
+#ifdef DDRAW_DEBUG
+ fprintf(stderr, "New refresh rate for %d bpp: %dx%d at %d Hz\n", (bpp+1)*8, (int)desc->dwWidth, (int)desc->dwHeight, refreshRate);
+#endif
+				}
+				break;
+			}
 			++SDL_nummodes[bpp];
 			enumrect = (struct DX5EnumRect*)malloc(sizeof(struct DX5EnumRect));
 			if ( !enumrect ) {
 				SDL_OutOfMemory();
 				return(DDENUMRET_CANCEL);
 			}
+			enumrect->refreshRate = refreshRate;
 			enumrect->r.x = 0;
 			enumrect->r.y = 0;
 			enumrect->r.w = (Uint16)desc->dwWidth;
 			enumrect->r.h = (Uint16)desc->dwHeight;
 			enumrect->next = enumlists[bpp];
 			enumlists[bpp] = enumrect;
+#ifdef DDRAW_DEBUG
+ fprintf(stderr, "New mode for %d bpp: %dx%d at %d Hz\n", (bpp+1)*8, (int)desc->dwWidth, (int)desc->dwHeight, refreshRate);
+#endif
 			break;
 	}
 
-
 	return(DDENUMRET_OK);
 }
 
 void SetDDerror(const char *function, int code)
 {
 	static char *error;
-	static char  errbuf[BUFSIZ];
+	static char  errbuf[1024];
 
 	errbuf[0] = 0;
 	switch (code) {
@@ -900,7 +930,7 @@ int DX5_VideoInit(_THIS, SDL_PixelFormat *vformat)
 	for ( i=0; i<NUM_MODELISTS; ++i )
 		enumlists[i] = NULL;
 
-	result = IDirectDraw2_EnumDisplayModes(ddraw2,0,NULL,this,EnumModes2);
+	result = IDirectDraw2_EnumDisplayModes(ddraw2,DDEDM_REFRESHRATES,NULL,this,EnumModes2);
 	if ( result != DD_OK ) {
 		SetDDerror("DirectDraw2::EnumDisplayModes", result);
 		return(-1);
@@ -914,7 +944,7 @@ int DX5_VideoInit(_THIS, SDL_PixelFormat *vformat)
 			return(-1);
 		}
 		for ( j = 0, rect = enumlists[i]; rect; ++j, rect = rect->next ) {
-			SDL_modelist[i][j]=(SDL_Rect *)rect;
+			SDL_modelist[i][j] = &rect->r;
 		}
 		SDL_modelist[i][j] = NULL;
 	}
@@ -973,17 +1003,14 @@ SDL_Surface *DX5_SetVideoMode(_THIS, SDL_Surface *current,
 			(WS_OVERLAPPED|WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX);
 	const DWORD resizestyle =
 			(WS_THICKFRAME|WS_MAXIMIZEBOX);
+	int windowX, windowY;
 	DDSURFACEDESC ddsd;
 	LPDIRECTDRAWSURFACE  dd_surface1;
 	LPDIRECTDRAWSURFACE3 dd_surface3;
-	BOOL was_visible;
 
 #ifdef DDRAW_DEBUG
  fprintf(stderr, "Setting %dx%dx%d video mode\n", width, height, bpp);
 #endif
-	/* See whether or not we should center the window */
-	was_visible = IsWindowVisible(SDL_Window);
-
 	/* Clean up any previous DirectDraw surfaces */
 	if ( current->hwdata ) {
 		this->FreeHWSurface(this, current);
@@ -994,11 +1021,13 @@ SDL_Surface *DX5_SetVideoMode(_THIS, SDL_Surface *current,
 		SDL_primary = NULL;
 	}
 
+#ifndef NO_CHANGEDISPLAYSETTINGS
 	/* Unset any previous OpenGL fullscreen mode */
 	if ( (current->flags & (SDL_OPENGL|SDL_FULLSCREEN)) ==
 	                       (SDL_OPENGL|SDL_FULLSCREEN) ) {
 		ChangeDisplaySettings(NULL, 0);
 	}
+#endif
 
 	/* Clean up any GL context that may be hanging around */
 	if ( current->flags & SDL_OPENGL ) {
@@ -1057,6 +1086,7 @@ SDL_Surface *DX5_SetVideoMode(_THIS, SDL_Surface *current,
 		video->h = height;
 		video->pitch = SDL_CalculatePitch(video);
 
+#ifndef NO_CHANGEDISPLAYSETTINGS
 		/* Set fullscreen mode if appropriate.
 		   Ugh, since our list of valid video modes comes from
 		   the DirectX driver, we may not actually be able to
@@ -1077,10 +1107,11 @@ SDL_Surface *DX5_SetVideoMode(_THIS, SDL_Surface *current,
 				SDL_fullscreen_mode = settings;
 			}
 		}
+#endif /* !NO_CHANGEDISPLAYSETTINGS */
 
 		style = GetWindowLong(SDL_Window, GWL_STYLE);
 		style &= ~(resizestyle|WS_MAXIMIZE);
-		if ( (video->flags & SDL_FULLSCREEN) == SDL_FULLSCREEN ) {
+		if ( video->flags & SDL_FULLSCREEN ) {
 			style &= ~windowstyle;
 			style |= directstyle;
 		} else {
@@ -1096,32 +1127,66 @@ SDL_Surface *DX5_SetVideoMode(_THIS, SDL_Surface *current,
 					video->flags |= SDL_RESIZABLE;
 				}
 			}
+#if WS_MAXIMIZE
 			if (IsZoomed(SDL_Window)) style |= WS_MAXIMIZE;
+#endif
 		}
-		SetWindowLong(SDL_Window, GWL_STYLE, style);
+
+		/* DJM: Don't piss of anyone who has setup his own window */
+		if ( SDL_windowid == NULL )
+			SetWindowLong(SDL_Window, GWL_STYLE, style);
 
 		/* Resize the window (copied from SDL WinDIB driver) */
 		if ( SDL_windowid == NULL ) {
+			HWND top;
 			UINT swp_flags;
+			const char *window = getenv("SDL_VIDEO_WINDOW_POS");
+			const char *center = getenv("SDL_VIDEO_CENTERED");
+
+			if ( !SDL_windowX && !SDL_windowY ) {
+				if ( window ) {
+					if ( sscanf(window, "%d,%d", &x, &y) == 2 ) {
+						SDL_windowX = x;
+						SDL_windowY = y;
+					}
+					if ( strcmp(window, "center") == 0 ) {
+						center = window;
+						window = NULL;
+					}
+				}
+			}
+			swp_flags = (SWP_NOCOPYBITS | SWP_SHOWWINDOW);
 
 			SDL_resizing = 1;
-			bounds.left = 0;
-			bounds.top = 0;
-			bounds.right = video->w;
-			bounds.bottom = video->h;
-			AdjustWindowRect(&bounds, GetWindowLong(SDL_Window, GWL_STYLE), FALSE);
+			bounds.left = SDL_windowX;
+			bounds.top = SDL_windowY;
+			bounds.right = SDL_windowX+video->w;
+			bounds.bottom = SDL_windowY+video->h;
+			AdjustWindowRectEx(&bounds, GetWindowLong(SDL_Window, GWL_STYLE), FALSE, 0);
 			width = bounds.right-bounds.left;
 			height = bounds.bottom-bounds.top;
-			x = (GetSystemMetrics(SM_CXSCREEN)-width)/2;
-			y = (GetSystemMetrics(SM_CYSCREEN)-height)/2;
+			if ( (flags & SDL_FULLSCREEN) ) {
+				x = (GetSystemMetrics(SM_CXSCREEN)-width)/2;
+				y = (GetSystemMetrics(SM_CYSCREEN)-height)/2;
+			} else if ( SDL_windowX || SDL_windowY || window ) {
+				x = bounds.left;
+				y = bounds.top;
+			} else if ( center ) {
+				x = (GetSystemMetrics(SM_CXSCREEN)-width)/2;
+				y = (GetSystemMetrics(SM_CYSCREEN)-height)/2;
+			} else {
+				x = y = -1;
+				swp_flags |= SWP_NOMOVE;
+			}
 			if ( y < 0 ) { /* Cover up title bar for more client area */
 				y -= GetSystemMetrics(SM_CYCAPTION)/2;
 			}
-			swp_flags = (SWP_NOCOPYBITS | SWP_NOZORDER | SWP_SHOWWINDOW);
-			if ( was_visible && !(flags & SDL_FULLSCREEN) ) {
-				swp_flags |= SWP_NOMOVE;
+			if ( flags & SDL_FULLSCREEN ) {
+				top = HWND_TOPMOST;
+			} else {
+				top = HWND_NOTOPMOST;
 			}
-			SetWindowPos(SDL_Window, NULL, x, y, width, height, swp_flags);
+			SetWindowPos(SDL_Window, top, x, y, width, height, swp_flags);
 			SDL_resizing = 0;
 			SetForegroundWindow(SDL_Window);
 		}
@@ -1135,6 +1200,8 @@ SDL_Surface *DX5_SetVideoMode(_THIS, SDL_Surface *current,
 	}
 
 	/* Set the appropriate window style */
+	windowX = SDL_windowX;
+	windowY = SDL_windowY;
 	style = GetWindowLong(SDL_Window, GWL_STYLE);
 	style &= ~(resizestyle|WS_MAXIMIZE);
 	if ( (flags & SDL_FULLSCREEN) == SDL_FULLSCREEN ) {
@@ -1151,9 +1218,13 @@ SDL_Surface *DX5_SetVideoMode(_THIS, SDL_Surface *current,
 				style |= resizestyle;
 			}
 		}
+#if WS_MAXIMIZE
 		if (IsZoomed(SDL_Window)) style |= WS_MAXIMIZE;
+#endif
 	}
-	SetWindowLong(SDL_Window, GWL_STYLE, style);
+	/* DJM: Don't piss of anyone who has setup his own window */
+	if ( SDL_windowid == NULL )
+		SetWindowLong(SDL_Window, GWL_STYLE, style);
 
 	/* Set DirectDraw sharing mode.. exclusive when fullscreen */
 	if ( (flags & SDL_FULLSCREEN) == SDL_FULLSCREEN ) {
@@ -1166,27 +1237,53 @@ SDL_Surface *DX5_SetVideoMode(_THIS, SDL_Surface *current,
 		SetDDerror("DirectDraw2::SetCooperativeLevel", result);
 		return(NULL);
 	}
+	SDL_windowX = windowX;
+	SDL_windowY = windowY;
 
 	/* Set the display mode, if we are in fullscreen mode */
 	if ( (flags & SDL_FULLSCREEN) == SDL_FULLSCREEN ) {
+		RECT bounds;
+		struct DX5EnumRect *rect;
+		int maxRefreshRate;
+
 		/* Cover up desktop during mode change */
 		SDL_resizing = 1;
-		SetWindowPos(SDL_Window, NULL, 0, 0, 
-			GetSystemMetrics(SM_CXSCREEN),
-			GetSystemMetrics(SM_CYSCREEN),
-			(SWP_NOCOPYBITS | SWP_NOZORDER));
+		bounds.left = 0;
+		bounds.top = 0;
+		bounds.right = GetSystemMetrics(SM_CXSCREEN);
+		bounds.bottom = GetSystemMetrics(SM_CYSCREEN);
+		AdjustWindowRectEx(&bounds, GetWindowLong(SDL_Window, GWL_STYLE), FALSE, 0);
+		SetWindowPos(SDL_Window, HWND_TOPMOST,
+			bounds.left, bounds.top, 
+			bounds.right - bounds.left,
+			bounds.bottom - bounds.top, SWP_NOCOPYBITS);
 		SDL_resizing = 0;
 		ShowWindow(SDL_Window, SW_SHOW);
 		while ( GetForegroundWindow() != SDL_Window ) {
 			SetForegroundWindow(SDL_Window);
 			SDL_Delay(100);
 		}
-		result = IDirectDraw2_SetDisplayMode(ddraw2, width, height,
-								bpp, 0, 0);
+
+		/* find maximum monitor refresh rate for this resolution */
+		/* Dmitry Yakimov ftech@tula.net */
+		maxRefreshRate = 0; /* system default */
+		for ( rect = enumlists[bpp / 8 - 1]; rect; rect = rect->next ) {
+			if ( (width == rect->r.w) && (height == rect->r.h) ) {
+				maxRefreshRate = rect->refreshRate;
+				break;
+			}
+		}
+#ifdef DDRAW_DEBUG
+ fprintf(stderr, "refresh rate = %d Hz\n", maxRefreshRate);
+#endif
+
+		result = IDirectDraw2_SetDisplayMode(ddraw2, width, height, bpp, maxRefreshRate, 0);
 		if ( result != DD_OK ) {
-			/* We couldn't set fullscreen mode, try window */
-			return(DX5_SetVideoMode(this, current,
-				width, height, bpp, flags & ~SDL_FULLSCREEN)); 
+			result = IDirectDraw2_SetDisplayMode(ddraw2, width, height, bpp, 0, 0);
+			if ( result != DD_OK ) {
+				/* We couldn't set fullscreen mode, try window */
+				return(DX5_SetVideoMode(this, current, width, height, bpp, flags & ~SDL_FULLSCREEN)); 
+			}
 		}
 		DX5_DInputReset(this, 1);
 	} else {
@@ -1368,7 +1465,15 @@ SDL_Surface *DX5_SetVideoMode(_THIS, SDL_Surface *current,
 			}
 		}
 		dd_surface3 = NULL;
+#if 0 /* FIXME: enable this when SDL consistently reports lost surfaces */
+		if ( (flags & SDL_HWSURFACE) == SDL_HWSURFACE ) {
+			video->flags |= SDL_HWSURFACE;
+		} else {
+			video->flags |= SDL_SWSURFACE;
+		}
+#else
 		video->flags |= SDL_SWSURFACE;
+#endif
 		if ( (flags & SDL_RESIZABLE) && !(flags & SDL_NOFRAME) ) {
 			video->flags |= SDL_RESIZABLE;
 		}
@@ -1415,6 +1520,8 @@ SDL_Surface *DX5_SetVideoMode(_THIS, SDL_Surface *current,
 		RECT bounds;
 		int  x, y;
 		UINT swp_flags;
+		const char *window = getenv("SDL_VIDEO_WINDOW_POS");
+		const char *center = getenv("SDL_VIDEO_CENTERED");
 
 		/* Create and set a clipper on our primary surface */
 		if ( SDL_clipper == NULL ) {
@@ -1446,27 +1553,47 @@ SDL_Surface *DX5_SetVideoMode(_THIS, SDL_Surface *current,
 			return(NULL);
 		}
 
-		/* Set the size of the window, centering and adjusting */
+		if ( !SDL_windowX && !SDL_windowY ) {
+			if ( window ) {
+				if ( sscanf(window, "%d,%d", &x, &y) == 2 ) {
+					SDL_windowX = x;
+					SDL_windowY = y;
+				}
+				if ( strcmp(window, "center") == 0 ) {
+					center = window;
+					window = NULL;
+				}
+			}
+		}
+		swp_flags = SWP_NOCOPYBITS;
+
 		SDL_resizing = 1;
-		bounds.left = 0;
-		bounds.top = 0;
-		bounds.right = video->w;
-		bounds.bottom = video->h;
-		AdjustWindowRect(&bounds, GetWindowLong(SDL_Window, GWL_STYLE),
-									FALSE);
+		bounds.left = SDL_windowX;
+		bounds.top = SDL_windowY;
+		bounds.right = SDL_windowX+video->w;
+		bounds.bottom = SDL_windowY+video->h;
+		AdjustWindowRectEx(&bounds, GetWindowLong(SDL_Window, GWL_STYLE), FALSE, 0);
 		width = bounds.right-bounds.left;
 		height = bounds.bottom-bounds.top;
-		x = (GetSystemMetrics(SM_CXSCREEN)-width)/2;
-		y = (GetSystemMetrics(SM_CYSCREEN)-height)/2;
+		if ( (flags & SDL_FULLSCREEN) ) {
+			x = (GetSystemMetrics(SM_CXSCREEN)-width)/2;
+			y = (GetSystemMetrics(SM_CYSCREEN)-height)/2;
+		} else if ( SDL_windowX || SDL_windowY || window ) {
+			x = bounds.left;
+			y = bounds.top;
+		} else if ( center ) {
+			x = (GetSystemMetrics(SM_CXSCREEN)-width)/2;
+			y = (GetSystemMetrics(SM_CYSCREEN)-height)/2;
+		} else {
+			x = y = -1;
+			swp_flags |= SWP_NOMOVE;
+		}
 		if ( y < 0 ) { /* Cover up title bar for more client area */
 			y -= GetSystemMetrics(SM_CYCAPTION)/2;
 		}
-		swp_flags = (SWP_NOCOPYBITS | SWP_NOZORDER);
-		if ( was_visible ) {
-			swp_flags |= SWP_NOMOVE;
-		}
-		SetWindowPos(SDL_Window, NULL, x, y, width, height, swp_flags);
+		SetWindowPos(SDL_Window, HWND_NOTOPMOST, x, y, width, height, swp_flags);
 		SDL_resizing = 0;
+
 	}
 	ShowWindow(SDL_Window, SW_SHOW);
 	SetForegroundWindow(SDL_Window);
@@ -1587,7 +1714,7 @@ static int DX5_AllocDDSurface(_THIS, SDL_Surface *surface,
 	memset(&ddsd, 0, sizeof(ddsd));
 	ddsd.dwSize = sizeof(ddsd);
 	result = IDirectDrawSurface3_Lock(dd_surface3, NULL,
-					&ddsd, DDLOCK_NOSYSLOCK, NULL);
+		&ddsd, (DDLOCK_NOSYSLOCK|DDLOCK_WAIT), NULL);
 	if ( result != DD_OK ) {
 		SetDDerror("DirectDrawSurface3::Lock", result);
 		goto error_end;
@@ -1731,10 +1858,10 @@ static int DX5_HWAccelBlit(SDL_Surface *src, SDL_Rect *srcrect,
 	/* Set it up.. the desination must have a DDRAW surface */
 	src_surface = src->hwdata->dd_writebuf;
 	dst_surface = dst->hwdata->dd_writebuf;
-	rect.top    = srcrect->y;
-	rect.bottom = srcrect->y+srcrect->h;
-	rect.left   = srcrect->x;
-	rect.right  = srcrect->x+srcrect->w;
+	rect.top    = (LONG)srcrect->y;
+	rect.bottom = (LONG)srcrect->y+srcrect->h;
+	rect.left   = (LONG)srcrect->x;
+	rect.right  = (LONG)srcrect->x+srcrect->w;
 	if ( (src->flags & SDL_SRCCOLORKEY) == SDL_SRCCOLORKEY )
 		flags = DDBLTFAST_SRCCOLORKEY;
 	else
@@ -1824,10 +1951,10 @@ static int DX5_FillHWRect(_THIS, SDL_Surface *dst, SDL_Rect *dstrect, Uint32 col
  fprintf(stderr, "HW accelerated fill at (%d,%d)\n", dstrect->x, dstrect->y);
 #endif
 	dst_surface = dst->hwdata->dd_writebuf;
-	area.top = dstrect->y;
-	area.bottom = dstrect->y+dstrect->h;
-	area.left = dstrect->x;
-	area.right = dstrect->x+dstrect->w;
+	area.top    = (LONG)dstrect->y;
+	area.bottom = (LONG)dstrect->y+dstrect->h;
+	area.left   = (LONG)dstrect->x;
+	area.right  = (LONG)dstrect->x+dstrect->w;
 	bltfx.dwSize = sizeof(bltfx);
 #if defined(NONAMELESSUNION)
 	bltfx.u5.dwFillColor = color;
@@ -1920,11 +2047,17 @@ static int DX5_FlipHWSurface(_THIS, SDL_Surface *surface)
 	LPDIRECTDRAWSURFACE3 dd_surface;
 
 	dd_surface = surface->hwdata->dd_surface;
+
+	/* to prevent big slowdown on fast computers, wait here instead of driver ring 0 code */
+	/* Dmitry Yakimov (ftech@tula.net) */
+	while(IDirectDrawSurface3_GetFlipStatus(dd_surface, DDGBS_ISBLTDONE) == DDERR_WASSTILLDRAWING);
+
 	result = IDirectDrawSurface3_Flip(dd_surface, NULL, DDFLIP_WAIT);
 	if ( result == DDERR_SURFACELOST ) {
 		result = IDirectDrawSurface3_Restore(
 						surface->hwdata->dd_surface);
-		result = IDirectDrawSurface3_Flip(dd_surface,NULL,DDFLIP_WAIT);
+		while(IDirectDrawSurface3_GetFlipStatus(dd_surface, DDGBS_ISBLTDONE) == DDERR_WASSTILLDRAWING);
+		result = IDirectDrawSurface3_Flip(dd_surface, NULL, DDFLIP_WAIT);
 	}
 	if ( result != DD_OK ) {
 		SetDDerror("DirectDrawSurface3::Flip", result);
@@ -1951,14 +2084,14 @@ void DX5_WindowUpdate(_THIS, int numrects, SDL_Rect *rects)
 	RECT src, dst;
 
 	for ( i=0; i<numrects; ++i ) {
-		src.top = rects[i].y;
-		src.bottom = rects[i].y+rects[i].h;
-		src.left = rects[i].x;
-		src.right = rects[i].x+rects[i].w;
-		dst.top = SDL_bounds.top+src.top;
-		dst.left = SDL_bounds.left+src.left;
+		src.top    = (LONG)rects[i].y;
+		src.bottom = (LONG)rects[i].y+rects[i].h;
+		src.left   = (LONG)rects[i].x;
+		src.right  = (LONG)rects[i].x+rects[i].w;
+		dst.top    = SDL_bounds.top+src.top;
+		dst.left   = SDL_bounds.left+src.left;
 		dst.bottom = SDL_bounds.top+src.bottom;
-		dst.right = SDL_bounds.left+src.right;
+		dst.right  = SDL_bounds.left+src.right;
 		result = IDirectDrawSurface3_Blt(SDL_primary, &dst, 
 					this->screen->hwdata->dd_surface, &src,
 							DDBLT_WAIT, NULL);
@@ -2205,16 +2338,29 @@ static int DX5_GetGammaRamp(_THIS, Uint16 *ramp)
 #endif /* !IDirectDrawGammaControl_SetGammaRamp */
 }
 
+static void FlushMessageQueue()
+{
+	MSG  msg;
+	while ( PeekMessage(&msg, NULL, 0, 0, PM_REMOVE) ) {
+		if ( msg.message == WM_QUIT ) break;
+		TranslateMessage( &msg );
+		DispatchMessage( &msg );
+	}
+}
+
 void DX5_VideoQuit(_THIS)
 {
 	int i, j;
 
 	/* If we're fullscreen GL, we need to reset the display */
 	if ( this->screen != NULL ) {
+#ifndef NO_CHANGEDISPLAYSETTINGS
 		if ( (this->screen->flags & (SDL_OPENGL|SDL_FULLSCREEN)) ==
 		                            (SDL_OPENGL|SDL_FULLSCREEN) ) {
 			ChangeDisplaySettings(NULL, 0);
+			ShowWindow(SDL_Window, SW_HIDE);
 		}
+#endif
 		if ( this->screen->flags & SDL_OPENGL ) {
 			WIN_GL_ShutDown(this);
 		}
@@ -2245,6 +2391,7 @@ void DX5_VideoQuit(_THIS)
 	DIB_QuitGamma(this);
 	if ( SDL_Window ) {
 		DX5_DestroyWindow(this);
+		FlushMessageQueue();
 	}
 
 	/* Free our window icon */
